
/*   
    A class for making "console windows", which can be used for doing basic
    character-oriented input/output.  Methods are provided for reading and
    writing the primitive data types plus Strings.
    
    Problems:  (1) This works with CodeWarrior on the Mac, but when
                   I tried it on a Sun under AppletViewer, I found that
                   the upperleft corner of the drawing area of a Frame
                   is not (0,0)!  Looks like it's at the upper left of the
                   window's border instead.  Is this device independence???

    I wrote this for use in an intro programming course as a kind of fudge,
    because System.in doesn't work on the Mac.  (And because neither InputStream
    nor any other standard input stream has any decent input methods anyway.)
    
    David Eck
    Department of Mathematics and Computer Science
    Hobart and William Smith Colleges
    Geneva, NY 14456
    eck@hws.edu
    
    June 30, 1996
    
    Modified August 27, 1996 to use an off-screeen canvas to hold a copy of 
         the stuff that's on the screen.
         
    Modified September 14, 1996 to keep the console window open after
         the program ends, until the user hits some key.

*/

   
import java.awt.*;

public class Console extends Frame {

   // ***************************** Constructors *******************************
   
   public Console() {   // default constructor just provides default window title and size
      this("Java I/O Console",25,80);
   }
   
   public Console(String title) {  // open window with specified title and default size
      this(title,25,80);
   }
   
   public Console(String title, int R, int C) {
              // create and open console window with given title and size.
              // R is the number of rows of text to be displayed in the window;
              // C is the number of columns of text

      super(title);
      
      rows = R;
      cols = C;

      setBackground(Color.white);
      setForeground(Color.black);
      font = new Font("Courier",Font.PLAIN,12);   // want monospaced font
      setFont(font);
      FontMetrics fm = getFontMetrics(font);

      lineHeight = fm.getHeight();   // initialize instance variables
      baseOffset = fm.getAscent();
      charWidth = fm.charWidth('W');
      contents = new char[rows][cols];
      lineLength = new int[rows];

      resize(2*border + cols * charWidth, border + rows * lineHeight);
      setResizable(false);
      
      show();

      makeOSC();  // make an off-screen canvas for double buffering.
                  //   (N.B.:  This wouldn't work if I did it before calling show() !!
      
   }
   
   public void close() {  // close the window; can't use the window after doing this!
      putln();
      put("   *** Console ready to close; Hit any key ***");
      closing = true;
   }
   
   public void clear() {  // clear the console window, return cursor to origin
      clearWin();  // (private method, defined below)
   }
   
   // *************************** I/O Methods *********************************
   
         // Methods for writing the primitive types, plus type String,
         // to the console window, with no extra spaces.

   public void put(int x)     { put(x,0); }   // Note: also handles byte and short!
   public void put(long x)    { put(x,0); }
   public void put(float x)   { put(x,0); }
   public void put(double x)  { put(x,0); }
   public void put(char x)    { put(x,0); }
   public void put(boolean x) { put(x,0); }
   public void put(String x)  { put(x,0); }


         // Methods for writing the primitive types, plus type String,
         // to the console window,followed by a carriage return, with
         // no extra spaces.

   public void putln(int x)      { put(x,0); newLine(); }  // Note: also handles byte and short!
   public void putln(long x)     { put(x,0); newLine(); }
   public void putln(float x)    { put(x,0); newLine(); }
   public void putln(double x)   { put(x,0); newLine(); }
   public void putln(char x)     { put(x,0); newLine(); }
   public void putln(boolean x)  { put(x,0); newLine(); }
   public void putln(String x)   { put(x,0); newLine(); }
  

         // Methods for writing the primitive types, plus type String,
         // to the console window, with a minimum field width of w,
         // and followed by a carriage  return.
         // If outut value is less than w characters, it is padded
         // with extra spaces in front of the value.

   public void putln(int x, int w)     { put(x,w); newLine(); }   // Note: also handles byte and short!
   public void putln(long x, int w)    { put(x,w); newLine(); }
   public void putln(float x, int w)   { put(x,w); newLine(); }
   public void putln(double x, int w)  { put(x,w); newLine(); }
   public void putln(char x, int w)    { put(x,w); newLine(); }
   public void putln(boolean x, int w) { put(x,w); newLine(); }
   public void putln(String x, int w)  { put(x,w); newLine(); }


          // Method for outputting a carriage return

   public void putln() { newLine(); }
   

         // Methods for writing the primitive types, plus type String,
         // to the console window, with minimum field width w.
   
   public void put(int x, int w)     { dumpString(String.valueOf(x), w); }   // Note: also handles byte and short!
   public void put(long x, int w)    { dumpString(String.valueOf(x), w); }
   public void put(float x, int w)   { dumpString(String.valueOf(x), w); }
   public void put(double x, int w)  { dumpString(String.valueOf(x), w); }
   public void put(char x, int w)    { dumpString(String.valueOf(x), w); }
   public void put(boolean x, int w) { dumpString(String.valueOf(x), w); }
   public void put(String x, int w)  { dumpString(x, w); }
   
   
         // Methods for reading in the primitive types, plus "words" and "lines".
         // The "getln..." methods discard any extra input, up to the next carriage return.
         // A "word" read by getlnWord() is any sequence of non-blank characters.
         // A "line" read by getlnString() or getln() is everything up to next CR;
         //    the carriage return is not part of the returned value, but it is
         //    read and discarded.
         // Note that all input methods except those for char's and lines will
         // skip past any blanks and carriage returns to find a non-blank value.
         // getln() can return an empty string; getChar() and getlnChar() can 
         //    return a space or a linefeed ('\n') character.
         // Acceptable boolean values are the "words": true, false, t, f, yes,
         //    no, y, n, 0, or 1;  uppercase letters are OK.
         // None of these can produce an error; if an error is found in input,
         //    the user is forced to re-enter.
   
   public byte getlnByte()       { byte x=getByte();       emptyBuffer();  return x; }
   public short getlnShort()     { short x=getShort();     emptyBuffer();  return x; }
   public int getlnInt()         { int x=getInt();         emptyBuffer();  return x; }
   public long getlnLong()       { long x=getLong();       emptyBuffer();  return x; }
   public float getlnFloat()     { float x=getFloat();     emptyBuffer();  return x; }
   public double getlnDouble()   { double x=getDouble();   emptyBuffer();  return x; }
   public char getlnChar()       { char x=getChar();       emptyBuffer();  return x; }
   public boolean getlnBoolean() { boolean x=getBoolean(); emptyBuffer();  return x; }
   public String getlnWord()     { String x=getWord();     emptyBuffer();  return x; }
   public String getlnString()   { return getln(); }  // same as getln()
   public String getln() {
      StringBuffer s = new StringBuffer(100);
      char ch = readChar();
      while (ch != '\n') {
         s.append(ch);
         ch = readChar();
      }
      return s.toString();
   }
   
   
   public byte getByte()   { return (byte)readInteger(-128L,127L); }
   public short getShort() { return (short)readInteger(-32768L,32767L); }   
   public int getInt()     { return (int)readInteger((long)Integer.MIN_VALUE, (long)Integer.MAX_VALUE); }
   public long getLong()   { return readInteger(Long.MIN_VALUE, Long.MAX_VALUE); }
   
   public char getChar()   { return readChar(); }
   public char peek()      { return lookChar(); }
   
   public float getFloat() {  // can return positive or negative infinity
      float x = 0.0F;
      while (true) {
         String str = readRealString();
         if (str.equals("")) {
             StringBuffer s = new StringBuffer(34);
             int i=0;
             char ch=lookChar();
             while (i<30) {
                if (ch == '\n')
                   break;
                s.append(readChar());
                ch = lookChar();
                i++;
                if (i == 30)
                  s.append("...");
             }
             errorMessage("Illegal floating point input",
                          "Real number in the range " + Float.MIN_VALUE + " to " + Float.MAX_VALUE,
                          s.toString());
         }
         else {
            Float f = null;
            try { f = Float.valueOf(str); }
            catch (NumberFormatException e) {
               errorMessage("Illegal floating point input",
                            "Real number in the range " + Float.MIN_VALUE + " to " + Float.MAX_VALUE,
                            str);
               continue;
            }
            x = f.floatValue();
            break;
         }
      }
      return x;
   }
   
   public double getDouble() {  // can return positive or negative infinity
      double x = 0.0;
      while (true) {
         String str = readRealString();
         if (str.equals("")) {
             int i=0;
             StringBuffer s = new StringBuffer(34);
             char ch = lookChar();
             while (i<30) {
                if (ch == '\n')
                   break;
                s.append(readChar());
                ch = lookChar();
                i++;
                if (i == 30)
                  s.append("...");
             }
             errorMessage("Illegal floating point input",
                          "Real number in the range " + Double.MIN_VALUE + " to " + Double.MAX_VALUE,
                          s.toString());
         }
         else {
            Double f = null;
            try { f = Double.valueOf(str); }
            catch (NumberFormatException e) {
               errorMessage("Illegal floating point input",
                            "Real number in the range " + Double.MIN_VALUE + " to " + Double.MAX_VALUE,
                            str);
               continue;
            }
            x = f.doubleValue();
            break;
         }
      }
      return x;
   }
   
   public String getWord() {
      char ch = lookChar();
      while (ch == ' ' || ch == '\n') {
         readChar();
         ch = lookChar();
      }
      StringBuffer str = new StringBuffer(50);
      while (ch != ' ' && ch != '\n') {
         str.append(readChar());
         ch = lookChar();
      }
      return str.toString();
   }
   
   public boolean getBoolean() {
      boolean ans = false;
      while (true) {
         String s = getWord();
         if ( s.equalsIgnoreCase("true") || s.equalsIgnoreCase("t") ||
                 s.equalsIgnoreCase("yes")  || s.equalsIgnoreCase("y") ||
                 s.equals("1") ) {
              ans = true;
              break;
          }
          else if ( s.equalsIgnoreCase("false") || s.equalsIgnoreCase("f") ||
                 s.equalsIgnoreCase("no")  || s.equalsIgnoreCase("n") ||
                 s.equals("0") ) {
              ans = false;
              break;
          }
          else
             errorMessage("Illegal boolean input value.",
                          "one of:  true, false, t, f, yes, no, y, n, 0, or 1",
                          s);
      }
      return ans;
   }
   
   // ***************** Everything beyond this point is private *******************
   
   // ********************** Utility routines for input/output ********************
   
   private boolean closing = false;  // set to true when method close() is called;
            // window closes next time user presses a key
   
   
   private String readRealString() {   // read chars from input following syntax of real numbers
      StringBuffer s=new StringBuffer(50);
      char ch=lookChar();
      while (ch == ' ' || ch == '\n') {
          readChar();
          ch = lookChar();
      }
      if (ch == '-' || ch == '+') {
          s.append(readChar());
          ch = lookChar();
          while (ch == ' ') {
             readChar();
             ch = lookChar();
          }
      }
      while (ch >= '0' && ch <= '9') {
          s.append(readChar());
          ch = lookChar();
      }
      if (ch == '.') {
         s.append(readChar());
         ch = lookChar();
         while (ch >= '0' && ch <= '9') {
             s.append(readChar());
             ch = lookChar();
         }
      }
      if (ch == 'E' || ch == 'e') {
         s.append(readChar());
         ch = lookChar();
         if (ch == '-' || ch == '+') {
             s.append(readChar());
             ch = lookChar();
         }
         while (ch >= '0' && ch <= '9') {
             s.append(readChar());
             ch = lookChar();
         }
      }
      return s.toString();
   }

   private long readInteger(long min, long max) {  // read long integer, limited to specified range
      long x=0;
      while (true) {
         StringBuffer s=new StringBuffer(34);
         char ch=lookChar();
         while (ch == ' ' || ch == '\n') {
             readChar();
             ch = lookChar();
         }
         if (ch == '-' || ch == '+') {
             s.append(readChar());
             ch = lookChar();
             while (ch == ' ') {
                readChar();
                ch = lookChar();
             }
         }
         while (ch >= '0' && ch <= '9') {
             s.append(readChar());
             ch = lookChar();
         }
         if (s.equals("")){
             int i=0;
             while (i<30) {
                if (ch == '\n')
                   break;
                s.append(readChar());
                ch = lookChar();
                i++;
                if (i == 30)
                  s.append("...");
             }
             errorMessage("Illegal integer input",
                          "Integer in the range " + min + " to " + max,
                          s.toString());
         }
         else {
             String str = s.toString();
             try { 
                x = Long.parseLong(str);
             }
             catch (NumberFormatException e) {
                errorMessage("Illegal integer input",
                             "Integer in the range " + min + " to " + max,
                             str);
                continue;
             }
             break;
         }
      }
      return x;
   }
   
   private void dumpString(String str, int w) {   // output string to console
      for (int i=str.length(); i<w; i++)
         addChar(' ');
      for (int i=0; i<str.length(); i++)
         if ((int)str.charAt(i) >= 0x20 && (int)str.charAt(i) != 0x7F)  // no control chars or delete
            addChar(str.charAt(i));
         else if (str.charAt(i) == '\n' || str.charAt(i) == '\r')
            newLine();
   }
   
   private void errorMessage(String message, String expecting, String found) {
                  // inform user of error and force user to re-enter.
       if (currentCol > 0)
          newLine();
       newLine();
       dumpString("  *** Error in input: " + message + "\n", 0);
       dumpString("  ***      Expecting: " + expecting + "\n", 0);
       dumpString("  ***          Found: " + found + "\n\n", 0);
       dumpString("Please re-enter: ", 0);
       fillBuffer();  // N.B.:  effectively discards any extra stuff in buffer
   }
   
   // **************************** the input buffer ****************************
   
   private char[] buffer = new char[255];  // holds one line typed by user (carriage return is not stored)
                                           //   N.B.: limit of 255 chars on a line; system
                                           //         effectively inserts a CR if user exceeds this.
   private int bufferCount = 0;            // how many chars are in buffer
   private int bufferPos = 1;              // next char in buffer that has not yet been read
                                           //   N.B.: bufferCount == bufferPos is taken to
                                           //         refer to the CR at the end of the line.
   
   private boolean bufferEmpty() {  // anything left to read?
      return bufferPos > bufferCount;
   }
   
   private void emptyBuffer() {     // discard remaing input from buffer
      bufferPos = bufferCount + 1;
   }
   
   private void fillBuffer() {      // fill buffer with one line of input typed by user
        // N.B.:  data is transferred from typeahead buffer, where keys are actually
        //        stored as use types them
      bufferCount = 0;
      char ch = '\0';
      while (ch != '\n' && bufferCount < 255) {
         ch = getTypeAhead();  // get next typed character
         if (ch == 8) {  // char is a backspace; erase previous char if any
            if (bufferCount > 0) {
               bufferCount--;
               eraseChar();
            }
         }
         else if (ch != '\n') {  // add non-CR char to buffer; a CR ends the line but is not stored
            buffer[bufferCount] = ch;
            bufferCount++;
            addChar(ch);
         }
      }
      bufferPos = 0;
      newLine();
   }
   
   private char lookChar() {  // look at next char in line, without reading it
      if (bufferPos > bufferCount)
         fillBuffer();
      char ch = '\0';
      if (bufferPos == bufferCount)
         ch = '\n';
      else
         ch = buffer[bufferPos];
      return ch;
   }
   
   private char readChar() {  // read next char from line
      char ch = lookChar();
      bufferPos++;
      return ch;
   }
   
   // *************************** the typeahead buffer *************************
   
   private char[] typeahead = new char[255];  // holds chars typed by user that
                                              //   have not yet been processed
   private int typeaheadCount = 0;  // how many chars are currently in the typeahead
   
   synchronized private char getTypeAhead() {  
          // get a char from typeahead; wait for one if necessary
          // and blink cursor while waiting
      showCursor();
      while (typeaheadCount == 0) {  // wait for user to type
         try  { wait(250); }  // wake up at least every 1/4 second to blink cursor
         catch (InterruptedException e) { }
         if (typeaheadCount == 0) {
            if (cursorVisible)
               hideCursor();
            else
               showCursor();
         }
      }
      char ch = typeahead[0];
      for (int i=0; i<typeaheadCount; i++)
         typeahead[i] = typeahead[i+1];
      typeaheadCount--;
      try { Thread.sleep(25); }  // wait a bit (I'd like user to see what's going on)
      catch (InterruptedException e) { }
      if  (cursorVisible)
         hideCursor();
      return ch;
   }
   
   synchronized public boolean keyDown(Event evt, int key) {
      if (closing) {
         dispose();
         return true;
      }
      if (key == (int)'\r')   // treat return as linefeed
         key = (int)'\n';
      if (key == 0x7F)        // treate delete as backspace
         key = 8;
      if (key >= 0x20 || key == (int)'\n' || key == 8) {  // allow only printables, linefeed, backspace
         if (typeaheadCount < 255) {
            typeahead[typeaheadCount] = (char)key;
            typeaheadCount++;
            if (typeaheadCount == 1)  // getTypeAhead() might be waiting for this key,
               notify();              // so"notify" it
         }
      }
      return true;
   }
         

   // ************************** the visible window ****************************
   
   private int rows = 25;   // number of lines in console; could be reset in constructor
   private int cols = 80;   // number of chars per line
   
   private char[][] contents;   // characters in window (dimension is rows X cols)
   private int lineLength[];    // number of actual chars on each line (size is rows)

   private int currentRow = 0;  // chars are added and erased only at this position
   private int currentCol = 0;

   private int lineHeight; // total height of each line of text, including leading
   private int charWidth;  // column width
   private int baseOffset; // distance from top of line of text to baseline 
             // (N.B.:  font data must be set by constructor)
   
   private int border = 4;  // extra pixels around data in window
   
   private char[] fudgeChar = new char[1];  // Why isn't there a Gaphics.drawChar() routine????
   
   private boolean cursorVisible = false;   // Set to true if cursor is currently drawn on screen
   
   private Image OSC;  // off-screen canvas for double buffering; contains a copy of screen info
   private Graphics OSC_graphics;  // graphics context for OSC
   
   private Font font;  // font for console
   
   
   private void makeOSC() {
          // create an off-screen canvas for double buffering, if there is enough memory,
          // and make a graphics context for the OSC
      try {
         OSC = createImage(cols*charWidth,(rows+1)*lineHeight);
      }
      catch (Throwable e) {
         OSC = null;
      }
      if (OSC != null) {
         OSC_graphics = OSC.getGraphics();
         OSC_graphics.setColor(Color.white);
         OSC_graphics.fillRect(0,0,cols*charWidth,rows*lineHeight);
         OSC_graphics.setColor(Color.black);
         OSC_graphics.setFont(font);
      }
   }
      
   public synchronized void paint(Graphics g) {  // redraw all chars and cursor if necessary
      if (OSC != null) {
         g.drawImage(OSC,border,border,this);
      }
      else if (contents != null && lineLength != null) {
         for (int i=0; i<rows; i++)
            if (lineLength[i] > 0)
               for (int j=0; j<lineLength[i]; j++) {
                  fudgeChar[0] = contents[i][j];
                  g.drawChars(fudgeChar,0,1,border + j*charWidth,border + lineHeight*i + baseOffset);
               }
      }
      if (cursorVisible)
         putCursor(g);
   }
   
   private void putCursor(Graphics g) {  // draw cursor at current position
      g.drawLine(border + currentCol*charWidth, border + currentRow*lineHeight,
                 border + currentCol*charWidth, border + (currentRow+1)*lineHeight - 2);
   }
   
   private synchronized void hideCursor() {  // show cursor by drawing it in black
      if (cursorVisible) {
        cursorVisible = false;
        Graphics g=getGraphics();
        g.setColor(Color.white);
        putCursor(g);
        g.dispose();
      }
   }
   
   private synchronized void showCursor() {  // erase cursor by drawing it in white
      if (!cursorVisible) {
         cursorVisible = true;
         Graphics g=getGraphics();
         putCursor(g);
         g.dispose();
      }
   }
   
   private synchronized void clearWin() {  // remove all characters
       for (int i=0; i<rows; i++)
          lineLength[i] = 0;
       Graphics g = getGraphics();
       g.setColor(Color.white);
       g.fillRect(border,border,cols*charWidth,rows*lineHeight);
       currentRow = 0;
       currentCol = 0;
       g.dispose();
       if (cursorVisible)
          showCursor();
       if (OSC != null) {
          OSC_graphics.setColor(Color.white);
          OSC_graphics.fillRect(0,0,cols*charWidth,rows*lineHeight);
          OSC_graphics.setColor(Color.black);
       }
   }
   
   private synchronized void scrollUp() {  // scroll up by one line.
          // Both the on-screen and internal representations are scrolled.
          // N.B.:  If the off-screen canvas, OSC, exists, it is also scrolled;
          //        if the OSC is null, then window will not scroll properly
          //        when part of it is obscured.
      boolean needsCursor = cursorVisible;
      if (cursorVisible)
         hideCursor();
      for (int i = 0; i<rows-1; i++) {
         lineLength[i] = lineLength[i+1];
         for (int j=0; j<lineLength[i]; j++)
            contents[i][j] = contents[i+1][j];
      }
      lineLength[rows-1] = 0;
      if (OSC != null) {
         OSC_graphics.copyArea(0,lineHeight,cols*charWidth,(rows-1)*lineHeight,0,-lineHeight);
         OSC_graphics.setColor(Color.white);
         OSC_graphics.fillRect(0,(rows-1)*lineHeight,cols*charWidth,lineHeight);
         OSC_graphics.setColor(Color.black);
         Graphics g = getGraphics();
         g.drawImage(OSC,border,border,this);
         g.dispose();
      } 
      else {
         Graphics g = getGraphics();
         g.copyArea(border,border+lineHeight,cols*charWidth,(rows-1)*lineHeight,0,-lineHeight);
         g.setColor(Color.white);
         g.fillRect(border,border+(rows-1)*lineHeight,cols*charWidth,lineHeight);
         g.dispose();
      }
      if (needsCursor)
         showCursor();
   }
   
   private synchronized void newLine() {  // add a carriage return to widow and data stuctures
      boolean needsCursor = cursorVisible;
      if (cursorVisible)
         hideCursor();
      if (currentRow == rows-1)
         scrollUp();
      else
         currentRow++;
      currentCol = 0;    
      if (needsCursor)
         showCursor();
   }
   
   private synchronized void addChar(char ch) {  // add a character to window (assumed to be printable)
      boolean needsCursor = cursorVisible;
      if (cursorVisible)
         hideCursor();
      if (currentCol >= cols) {
         currentCol = 0;
         currentRow++;
         if (currentRow >= rows) {
            scrollUp();
            currentRow = rows-1;
         }
      }
      if (lineLength[currentRow] < currentCol) {
         for (int i=lineLength[currentRow]; i<currentCol; i++)
            contents[currentRow][i] = ' ';
         lineLength[currentRow] = currentCol;
      }
      contents[currentRow][currentCol] = ch;
      currentCol++;
      lineLength[currentRow]++;
      Graphics g = getGraphics();
      fudgeChar[0] = ch;
      g.drawChars(fudgeChar,0,1,border + (currentCol-1)*charWidth,border + lineHeight*currentRow + baseOffset);
      g.dispose();
      if (OSC != null)
         OSC_graphics.drawChars(fudgeChar,0,1,(currentCol-1)*charWidth,lineHeight*currentRow + baseOffset);
      if (needsCursor)
         showCursor();
   }
   
   private synchronized void eraseChar() {     // remove most recently added char or CR
      if (currentCol == 0 && currentRow == 0)
         return;
      boolean needsCursor = cursorVisible;
      if (cursorVisible)
         hideCursor();
      if (currentCol == 0) {
         currentRow--;
         currentCol = lineLength[currentRow];
      }
      else {
         currentCol--;
         lineLength[currentRow]--;
         Graphics g=getGraphics();
         fudgeChar[0]=contents[currentRow][currentCol];
         g.setColor(Color.white);
         g.drawChars(fudgeChar,0,1,border + currentCol*charWidth,border + lineHeight*currentRow + baseOffset);
         g.dispose();
         if (OSC != null) {
            OSC_graphics.setColor(Color.white);
            OSC_graphics.drawChars(fudgeChar,0,1,currentCol*charWidth,lineHeight*currentRow + baseOffset);
            OSC_graphics.setColor(Color.black);
         }
      }
      if (needsCursor)
         showCursor();
   }
   
} // end of class Console
